@@ -1,4 +1,4 @@
-;; Copyright (C) 2003 Shawn Betts
+;; Copyright (C) 2003-2008 Shawn Betts
 ;;
 ;;  This file is part of stumpwm.
 ;;
@@ -13,9 +13,8 @@
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with this software; see the file COPYING.  If not, write to
-;; the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
-;; Boston, MA 02111-1307 USA
+;; along with this software; see the file COPYING.  If not, see
+;; <http://www.gnu.org/licenses/>.
 
 ;; Commentary:
 ;;
@@ -24,7 +23,8 @@
 ;; Code:
 (in-package :stumpwm)
 
-(export '(*input-map*
+(export '(*input-history-ignore-duplicates*
+          *input-map*
 	  completing-read
 	  input-delete-region
 	  input-goto-char
@@ -37,7 +37,7 @@
 	  read-one-line))
 
 (defstruct input-line
-  string position history history-bk)
+  string position history history-bk password)
 
 (defvar *input-map* nil
   "This is the keymap containing all input editing key bindings.")
@@ -90,20 +90,16 @@
 (defvar *input-current-completions-idx* nil
   "The current index in the current completions list.")
 
+(defvar *input-history-ignore-duplicates* nil
+  "Do not add a command to the input history if it's already the first in the list.")
+
 ;;; keysym functions
 
-(defun is-modifier (keysym)
+(defun is-modifier (keycode)
   "Return t if keycode is a modifier"
-  ;; FIXME: This should depend on all the codes returned by ALL-MODIFIER-CODES
-  (let ((mods '("Mode_switch"
-                "Shift_L" "Shift_R"
-                "Control_L" "Control_R"
-                "Caps_Lock" "Shift_Lock"
-                "Meta_L" "Meta_R"
-                "Alt_L" "Alt_R"
-                "Super_L" "Super_R"
-                "Hyper_L" "Hyper_R")))
-    (member keysym (mapcar #'keysym-name->keysym mods))))
+  (or (find keycode *all-modifiers* :test 'eql)
+      ;; Treat No Symbol keys as modifiers (and therefore ignorable)
+      (= (xlib:keycode->keysym *display* keycode 0) 0)))
 
 (defun keycode->character (code mods)
   (let ((idx (if (member :shift mods) 1 0)))
@@ -117,24 +113,35 @@
                     (xlib:font-ascent (screen-font screen))))
          (win (screen-input-window screen)))
     ;; Window dimensions
-    (xlib:map-window win)
     (xlib:with-state (win)
       (setf (xlib:window-priority win) :above
             (xlib:drawable-height win) height))
+    (xlib:map-window win)
     ;; Draw the prompt
     (draw-input-bucket screen prompt input)
     ;; Ready to recieve input
-    (xlib:grab-keyboard (screen-input-window screen) :owner-p nil
-                        :sync-keyboard-p nil :sync-pointer-p nil)))
+    
+))
 
 (defun shutdown-input-window (screen)
   (xlib:ungrab-keyboard *display*)
   (xlib:unmap-window (screen-input-window screen)))
-
-(defun input-handle-key-press-event (&rest event-slots &key root code state &allow-other-keys)
+;; Hack to avoid clobbering input from numpads with numlock on.
+(defun input-handle-key-press-event (&rest event-slots 
+                                     &key event-key root code state 
+                                       &allow-other-keys)
   (declare (ignore event-slots root))
-  ;; FIXME: don't use a cons
-  (cons code state))
+  (let* ((numpad-map '((87 10 . 16) (88  11 . 16) (89 12 . 16) (106 61 . 16)
+                      (83 13 . 16) (84  14 . 16) (85 15 . 16) (86  21 . 17)
+                      (79 16 . 16) (80  17 . 16) (81 18 . 16) (63  17 . 17) 
+                      (82 20 . 16) (104 36 . 16) (91 60 . 16) (90  19 . 17)))
+         (numlock-on-p (= 2 (logand 2 (nth-value 4 (xlib:keyboard-control *display*)))))
+         (numpad-key (assoc code numpad-map)))
+    (when (and numlock-on-p numpad-key)
+      (setf code (first (rest numpad-key))
+            state (rest (rest numpad-key))))
+    ;; FIXME: don't use a cons
+    (list* event-key code state)))
 
 (defun input-handle-selection-event (&key window selection property &allow-other-keys)
   (declare (ignore selection))
@@ -145,18 +152,14 @@
 (defun read-key-handle-event (&rest event-slots &key display event-key &allow-other-keys)
   (declare (ignore display))
   (case event-key
-    (:key-release
-     't)
-    (:key-press
+    ((or :key-release :key-press)
      (apply 'input-handle-key-press-event event-slots))
     (t nil)))
 
 (defun read-key-or-selection-handle-event (&rest event-slots &key display event-key &allow-other-keys)
   (declare (ignore display))
   (case event-key
-    (:key-release
-     't)
-    (:key-press
+    ((or :key-release :key-press)
      (apply 'input-handle-key-press-event event-slots))
     (:selection-notify
      (apply 'input-handle-selection-event event-slots))
@@ -164,79 +167,98 @@
 
 (defun read-key ()
   "Return a dotted pair (code . state) key."
-  (do ((ret nil (xlib:process-event *display* :handler #'read-key-handle-event :timeout nil)))
-      ((consp ret) ret)))
+  (loop for ev = (xlib:process-event *display* :handler #'read-key-handle-event :timeout nil) do
+       (when (and (consp ev)
+                  (eq (first ev) :key-press))
+           (return (rest ev)))))
 
 (defun read-key-no-modifiers ()
   "Like read-key but never returns a modifier key."
-  (do ((k (read-key) (read-key)))
-      ((not (is-modifier (xlib:keycode->keysym *display* (car k) 0))) k)))
+  (loop for k = (read-key)
+       while (is-modifier (car k))
+       finally (return k)))
 
 (defun read-key-or-selection ()
-  (do ((ret nil (xlib:process-event *display* :handler #'read-key-or-selection-handle-event :timeout nil)))
-      ((or (stringp ret)
-           (consp ret))
-       ret)))
+  (loop for ev = (xlib:process-event *display* :handler #'read-key-or-selection-handle-event :timeout nil) do
+       (cond ((stringp ev)
+              (return ev))
+             ((and (consp ev)
+                   (eq (first ev) :key-press))
+              (return (rest ev))))))
 
 (defun make-input-string (initial-input)
   (make-array (length initial-input) :element-type 'character :initial-contents initial-input
               :adjustable t :fill-pointer t))
 
-(defun completing-read (screen prompt completions &optional (initial-input ""))
+(defun completing-read (screen prompt completions &key (initial-input "") require-match)
   "Read a line of input through stumpwm and return it with TAB
 completion. completions can be a list, an fbound symbol, or a
-function. if its an fbound symbol or a function then that
-function is passed the substring to complete on and is expected
-to return a list of matches."
+function. if its an fbound symbol or a function then that function is
+passed the substring to complete on and is expected to return a list
+of matches. If require-match argument is non-nil then the input must
+match with an element of the completions."
   (check-type completions (or list function symbol))
   (let ((*input-completions* completions)
         (*input-current-completions* nil)
         (*input-current-completions-idx* nil))
-    (read-one-line screen prompt initial-input)))
+    (let ((line (read-one-line screen prompt :initial-input initial-input :require-match require-match)))
+      (when line (string-trim " " line)))))
 
-(defun read-one-line (screen prompt &optional (initial-input ""))
+(defun read-one-line (screen prompt &key (initial-input "") require-match password)
   "Read a line of input through stumpwm and return it. returns nil if the user aborted."
   (let ((*input-last-command* nil)
         (input (make-input-line :string (make-input-string initial-input)
                                 :position (length initial-input)
-                                :history -1)))
-    (labels ((key-loop ()
+                                :history -1
+                                :password password)))
+    (labels ((match-input ()
+               (let* ((in (string-trim " " (input-line-string input)))
+                      (compls (input-find-completions in *input-completions*)))
+                 (and (consp compls)
+                      (string= in (car compls)))))
+             (key-loop ()
                (loop for key = (read-key-or-selection) do
                      (cond ((stringp key)
                             ;; handle selection
                             (input-insert-string input key)
                             (draw-input-bucket screen prompt input))
                            ;; skip modifiers
-                           ((is-modifier (xlib:keycode->keysym *display* (car key) 0)))
+                           ((is-modifier (car key)))
                            ((process-input screen prompt input (car key) (cdr key))
-                            (return (input-line-string input)))))))
+                            (if (or (not require-match)
+                                    (match-input))
+                                (return (input-line-string input))
+                                (draw-input-bucket screen prompt input "[No match]" t)))))))
       (setup-input-window screen prompt input)
       (catch :abort
         (unwind-protect
-             (key-loop)
+             (with-focus (screen-input-window screen)
+               (key-loop))
           (shutdown-input-window screen))))))
 
 (defun read-one-char (screen)
   "Read a single character from the user."
-  (grab-keyboard screen)
-  ;; FIXME: should this be in an unwind-protect to ungrab the kbd?
-  (prog1
-      (let ((k (read-key-no-modifiers)))
-        (keycode->character (car k) (xlib:make-state-keys (cdr k))))
-    (ungrab-keyboard)))
+  (with-focus (screen-key-window screen)
+    (let ((k (read-key-no-modifiers)))
+      (keycode->character (car k) (xlib:make-state-keys (cdr k))))))
 
-(defun draw-input-bucket (screen prompt input &optional errorp)
+
+(defun draw-input-bucket (screen prompt input &optional (tail "") errorp)
   "Draw to the screen's input window the contents of input."
   (let* ((gcontext (screen-message-gc screen))
          (win (screen-input-window screen))
          (prompt-width (xlib:text-width (screen-font screen) prompt :translate #'translate-id))
-         (string (input-line-string input))
+         (line-content (input-line-string input))
+         (string (if (input-line-password input)
+                     (make-string (length line-content) :initial-element #\*)
+                     line-content))
          (string-width (xlib:text-width (screen-font screen) string :translate #'translate-id))
-         (full-string-width (+ string-width
-                               (xlib:text-width (screen-font screen) " " :translate #'translate-id)))
+         (space-width  (xlib:text-width (screen-font screen) " "    :translate #'translate-id))
+         (tail-width   (xlib:text-width (screen-font screen) tail   :translate #'translate-id))
+         (full-string-width (+ string-width space-width))
          (pos (input-line-position input))
          (width (+ prompt-width
-                   (max 100 full-string-width))))
+                   (max 100 (+ full-string-width space-width tail-width)))))
     (xlib:with-state (win)
       (xlib:clear-area win :x (+ *message-window-padding*
                                  prompt-width
@@ -256,6 +278,12 @@ to return a list of matches."
                               string
                               :translate #'translate-id
                               :size 16)
+      (xlib:draw-image-glyphs win gcontext
+                              (+ *message-window-padding* prompt-width full-string-width space-width)
+                              (xlib:font-ascent (screen-font screen))
+                              tail
+                              :translate #'translate-id
+                              :size 16)
       ;; draw a block cursor
       (invert-rect screen win
                    (+ *message-window-padding*
@@ -277,9 +305,11 @@ to return a list of matches."
 
 (defun code-state->key (code state)
   (let* ((mods    (xlib:make-state-keys state))
-         (sym     (xlib:keycode->keysym *display* code 0))
-         (upsym   (xlib:keycode->keysym *display* code 1))
-         (shift-p (and (find :shift mods) t)))
+         (shift-p (and (find :shift mods) t))
+         (altgr-p (and (intersection (modifiers-altgr *modifiers*) mods) t))
+         (base    (if altgr-p 2 0))
+         (sym     (xlib:keycode->keysym *display* code base))
+         (upsym   (xlib:keycode->keysym *display* code (+ base 1))))
     ;; If a keysym has a shift modifier, then use the uppercase keysym
     ;; and remove remove the shift modifier.
     (make-key :keysym (if (and shift-p (not (eql sym upsym)))
@@ -313,9 +343,8 @@ to return a list of matches."
 position. @var{input} must be of type @var{input-line}. Input
 functions are passed this structure as their first argument."
   (check-type string string)
-  ;; FIXME: obviously this is substandard
-  (map nil (lambda (c) (input-insert-char input c))
-       string))
+  (loop for c across string
+	do (input-insert-char input c)))
 
 (defun input-point (input)
   "Return the position of the cursor."
@@ -521,7 +550,8 @@ functions are passed this structure as their first argument."
 
 (defun input-self-insert (input key)
   (let ((char (xlib:keysym->character *display* (key-keysym key))))
-    (if (or (key-mods-p key) (null char))
+    (if (or (key-mods-p key) (null char)
+            (not (characterp char)))
         :error
         (input-insert-char input char))))
 
@@ -551,13 +581,16 @@ input (pressing Return), nil otherwise."
                    :error))))
     (case (process-key code state)
       (:done
-       (push (input-line-string input) *input-history*)
+       (unless (or (input-line-password input)
+                   (and *input-history-ignore-duplicates*
+                        (string= (input-line-string input) (first *input-history*))))
+         (push (input-line-string input) *input-history*))
        :done)
       (:abort
        (throw :abort t))
       (:error
        ;; FIXME draw inverted text
-       (draw-input-bucket screen prompt input t)
+       (draw-input-bucket screen prompt input "" t)
        nil)
       (t
        (draw-input-bucket screen prompt input)
@@ -600,7 +633,9 @@ input (pressing Return), nil otherwise."
                          (find-mod "Hyper_R" codes))
                      (push mod (modifiers-hyper modifiers)))
                     ((find-mod "Num_Lock" codes)
-                     (push mod (modifiers-numlock modifiers)))))
+                     (push mod (modifiers-numlock modifiers)))
+                    ((find-mod "ISO_Level3_Shift" codes)
+                     (push mod (modifiers-altgr modifiers)))))
         ;; If alt is defined but meta isn't set meta to alt and clear alt
         (when (and (modifiers-alt modifiers)
                    (null (modifiers-meta modifiers)))
@@ -639,3 +674,20 @@ input (pressing Return), nil otherwise."
 
 ;; (defun cook-keycode (code state)
 ;;   (values (xlib:keycode->keysym *display* code 0) (x11mod->stumpmod state)))
+
+(defun y-or-n-p (message)
+  "ask a \"y or n\" question on the current screen and return T if the
+user presses 'y'"
+  (message "~a(y or n) " message)
+  (char= (read-one-char (current-screen))
+        #\y))
+
+(defun yes-or-no-p (message)
+  "ask a \"yes or no\" question on the current screen and return T if the
+user presses 'yes'"
+  (loop for line = (read-one-line (current-screen)
+                                  (format nil "~a(yes or no) " message))
+        until (find line '("yes" "no") :test 'string-equal)
+        do (message "Please answer yes or no")
+           (sleep 1)
+        finally (return (string-equal line "yes"))))
